// BeOS driver for w83781D/w83782D Wimbond H/W Monitoring IC
//
// References
// ----------
// 
// "W83781D Winbond H/W Monitoring IC Data Sheet", w83781d.pdf, nov 1997, revision 0.60
// "W83782D Winbond H/W Monitoring IC Data Sheet", w83782d.pdf, avril 1999, revision 0.55


#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>
#include <stdio.h>

#define DEBUG 1

#include <Debug.h>

/*
#ifdef DEBUG
	#define DEBUG_ONLY(expr)	{ expr; }
#else
	#define DEBUG_ONLY(expr)	NULL
#endif
*/

#include "Sensors.h"

#include "w8378xx.h"


#ifdef CODEWARRIOR
	#pragma mark [Local definitions]
#endif

// Local definitions
// -----------------


#ifndef DRIVER_NAME
	#define DRIVER_NAME 		"generic"
#endif
#define DEBUGGER_COMMAND	 	DRIVER_NAME

#define DEVICE_NAME 			DRIVER_NAME
#define DEVICE_NAME_LENGTH		(64)			
#define DEVICE_PATH 			"misc/sensor/" DEVICE_NAME "/"
#define DEVICE_MAX_ENTRIES 		(8)			

typedef struct device {
	long			id;
	const char *	name;
	uint32			flags;
	uint32			debug_level;

	// Locks mechanisms
	uint32			open_count;
	cpu_status		cs;					// Tempory storage for disable_interrupt()
	spinlock		hardware_spinlock;	// Hardware access *hardlock*
	sem_id			init_semaphore;		// Semaphore for init locking
	sem_id			update_semaphore;	// Semaphore for update locking
	bigtime_t		last_updated;		// Time we read values, 0 = never

	// Monitor datas
	uint32			nb_sensors;
	SENSOR *		sensors;
	void *			data;				// sensors chip specific datas	
	
	// Somes statistics
	volatile		nb_updates;
} device;

	
// globals for driver instances
static isa_module_info *	g_isa = NULL;
static char *				g_device_name_list[DEVICE_MAX_ENTRIES + 1] 	= { NULL };
static device *				g_device_list[DEVICE_MAX_ENTRIES + 1] 		= { NULL };

// Prototypes of device hooks functions
static status_t open_hook(const char *name, uint32 flags, void ** cookie);
static status_t close_hook(void * cookie);
static status_t free_hook(void * cookie);
static status_t control_hook(void * cookie, uint32 msg,void * buf, size_t len);
static status_t read_hook(void * cookie, off_t pos, void * buf, size_t * len);
static status_t write_hook(void * cookie, off_t pos, const void * buf, size_t * len);

/* And, so, the device hooks list */
static device_hooks g_device_hooks = {
	open_hook, 			/* -> open entry point */
	close_hook,         /* -> close entry point */
	free_hook,          /* -> free entry point */
	control_hook,       /* -> control entry point */
	read_hook,          /* -> read entry point */
	write_hook,         /* -> write entry point */
	NULL,               /* -> select entry point */
	NULL,               /* -> deselect entry point */
	NULL,               /* -> readv */
	NULL                /* -> writev */
};

// privates prototypes
// -------------------

static status_t 	device_find_supported(long * i, void ** detected_device_data);
static status_t 	device_reset(void * detected_device_data);
static status_t 	device_init(device * dev);
static status_t		device_uninit(device * dev);

static SENSOR *		device_find_sensor(device *	dev, char * name, int id);

static status_t 	w8378xx_update_values(device * dev);
static status_t 	w8378xx_init(device * dev);
static status_t 	w8378xx_uninit(device * dev);
static int			w8378xx_read(W8378xx * w, int reg);
static status_t		w8378xx_write(W8378xx * w, int reg, int value);

#ifdef DEBUGGER_COMMAND
	static int 		device_debugger_command(int argc, char ** argv);
#endif

static status_t 	device_clear_statistics(device * dev);

#ifdef CODEWARRIOR
	#pragma mark [Driver API calls]
#endif

// Driver API calls
// ----------------


_EXPORT int32 api_version = B_CUR_DRIVER_API_VERSION;


// -------------------------------------------------------
_EXPORT status_t init_hardware (void)
{
	status_t			status;
	long				nb;
	long				entry;
	void *				ddd;	// stand for "detected_device_data", an opaque *thing*

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- init_hardware()\n") );

	g_isa = NULL;
	g_device_list[0] = NULL;
	g_device_name_list[0] = NULL;

	nb = 0;
	entry = 0;
	ddd = NULL;
	while ( device_find_supported(&entry, &ddd) == B_OK )
		{
		entry++; // get ready for next one ?

		if (ddd == NULL )
			continue;

		device_reset(ddd);
		nb++;
		};
	
	if ( NULL != g_isa )
		put_module(B_ISA_MODULE_NAME);
	g_isa = NULL;
	
	if ( nb == 0 )
		DEBUG_ONLY( dprintf(DRIVER_NAME ": no supported device detected. Sorry.\n") ); 

	return ( nb > 0) ? B_OK : ENODEV ;
}



// -------------------------------------------------------
_EXPORT status_t init_driver(void)
{
	status_t	status;
	long		nb;
	long		entry;
	void *		ddd;	// stand for "detected_device_data", an opaque *thing*

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- init_driver(), built %s %s\n", __DATE__, __TIME__) );

#ifdef DEBUG
	if ( load_driver_symbols(DRIVER_NAME) != B_OK )
		dprintf(DRIVER_NAME ": load_driver_symbols(\"" DRIVER_NAME "\") FAILED !\n");
	else
		dprintf(DRIVER_NAME ": load_driver_symbols(\"" DRIVER_NAME "\") done.\n");
#endif

	status = get_module(B_ISA_MODULE_NAME, (module_info **) &g_isa);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": get_module(\"" B_ISA_MODULE_NAME "\") FAILED: %s\n",
							strerror(status)) );
		return status;
		};
		
	g_device_list[0] = NULL;

	nb = 0;
	entry = 0;
	ddd = NULL;
	while ( device_find_supported(&entry, &ddd) == B_OK )
		{
		char		name[ DEVICE_NAME_LENGTH + 1 ];
		device *	dev;

		entry++; // get ready for another one ?
		
		if ( NULL == ddd )
			// ARRGGGH, strange supported device detected !!!
			continue;

		sprintf(name, "%s%ld", DEVICE_PATH, nb);

		g_device_name_list[nb] 	= (char *) malloc(strlen(name) + 1);
		g_device_list[nb] 		= (device *) malloc(sizeof(device));

		// Check that mallocs succeed...
		if ( (NULL == g_device_name_list[nb]) ||
			 (NULL == g_device_list[nb])	)
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": malloc() FAILED -> ABORT !") );

			if ( NULL != g_device_name_list[nb] )
				free(g_device_name_list[nb]);
			if ( NULL != g_device_list[nb] )
				free(g_device_list[nb]);
			break;
			};

		// Okay, let's add a new device to global list
		strcpy(g_device_name_list[nb], name);

		dev = g_device_list[nb];
			
		memset(dev, 0, sizeof(*dev));

		dev->id	  			= nb;
		dev->name 			= g_device_name_list[nb];
		dev->flags			= 0;
		dev->open_count		= 0;

		dev->data			= ddd;	// opaque *thing* !!!
				
		dev->init_semaphore	= create_sem(0, DRIVER_NAME "_init_semaphore");	// 0 -> block on first acquire
		set_sem_owner(dev->init_semaphore, B_SYSTEM_TEAM); 	// Make kernel owner of init_semaphore

		DEBUG_ONLY( dprintf(DRIVER_NAME ": -> %s\n", dev->name) );

		nb++;	// Okay, one more device supported.
		if ( nb >= DEVICE_MAX_ENTRIES )
			break;	// sorry, no more device supported by this driver ! 
		};

	// Mark end of list
	g_device_list[nb] = NULL;
	g_device_name_list[nb] = NULL;

#ifdef DEBUGGER_COMMAND
	add_debugger_command(DEBUGGER_COMMAND, device_debugger_command, DRIVER_NAME " driver info");
#endif	
	return ( nb > 0) ? B_OK : ENODEV ;
}


// -------------------------------------------------------
_EXPORT void uninit_driver(void)
{
	long 	i; 

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- uninit_driver()\n") );

#ifdef DEBUGGER_COMMAND
	remove_debugger_command(DEBUGGER_COMMAND, device_debugger_command);
#endif

	// Free device list
	for ( i = 0; g_device_list[i]; i++ )
		{
		device * dev;
		
		dev = g_device_list[i];
		if ( dev->init_semaphore >= 0 )
			delete_sem(dev->init_semaphore);
		
		free(dev);
		};

	// Free device name list
	for ( i = 0; g_device_name_list[i]; i++ )
		free(g_device_name_list[i]);
	
	put_module(B_ISA_MODULE_NAME);
	g_isa = NULL;

}


// -------------------------------------------------------
_EXPORT const char ** publish_devices(void)
{
#ifdef DEBUG
	{
	long 	i;
	
	// Display published devices names
	dprintf(DRIVER_NAME ": ---- publish_devices():\n");

	for ( i = 0; g_device_name_list[i]; i++ )
		dprintf(DRIVER_NAME ": %s\n", g_device_name_list[i] );
	};
#endif

	return (const char **) g_device_name_list;
}


// -------------------------------------------------------
_EXPORT device_hooks * find_device
	(
	const char *	name
	)
{
	long 	i;
	char *	device_name;

	// Find device name
	for ( i = 0; (device_name = g_device_name_list[i]); i++ )
		{
		if ( strcmp(device_name, name) == 0 )
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- find_device: %s published, great !\n", name) );
			return &g_device_hooks;
			};
		};

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- find_device: %s NOT PUBLISHED, ARGH !\n", name) );
	return NULL; // We did'nt published this device name, it's weird, what's up, Kernel ?!?
}


#ifdef CODEWARRIOR
	#pragma mark [Device hooks]
#endif

// Device hooks
// ------------

// -------------------------------------------------------
static status_t open_hook
	(
	const char *	name, 
	uint32 			flags, 
	void **			cookie
	)
{
	device *	dev;
	long		i;
	status_t	status;

	// Find device name
	for ( i = 0; (dev = g_device_list[i]); i++ )
		{
		if ( strcmp(dev->name, name) == 0 )
			break;
		};

	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: %s NOT FOUND, ARGH !\n", name) );
		return EINVAL;
		};
		
	if ( atomic_add(&(dev->open_count), 1) == 0 )
		{
		// First open on this device :
		status = device_init(dev);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: device_init(%s)->%d, FAILURE !\n", dev->name, status) );
			return status;
			};
		}
	else
		// Not the first open on this device
		// But perhaps the first 'open' on this device is still pending,
		// so we wait for end (death !) of init_semaphore
		acquire_sem(dev->init_semaphore);

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: %s opened, great !\n", dev->name) );

	*cookie = dev;

	return B_OK; // B_ERROR;
}


// -------------------------------------------------------
static status_t close_hook
	(
	void *	cookie
	)
{
	device *	dev;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- close_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- close_hook %s\n", dev->name) );

	return B_OK;
}


// -------------------------------------------------------
static status_t free_hook
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- free_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- free_hook %s\n", dev->name) );

	if ( atomic_add(&(dev->open_count), -1) == 1 )
		// Last free() on this device :
		status = device_uninit(dev);
	else
		// not the last session, nothing to actually free
		status = B_OK;
	
	return status;
}



// -------------------------------------------------------
static status_t control_hook
	(
	void *	cookie, 
	uint32 	op_code, 
	void *	data, 
	size_t 	len
	)
{
	device *	dev;
	W8378xx *	w;

	dev = (device *) cookie;
	if ( (NULL == dev) || (dev->data == NULL) )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
		
	w = dev->data;
	
	switch ( op_code ) {
	case SENSOR_GET_MONITOR_INFO:
		{
		SENSORS_MONITOR	* 	sm;
		
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_GET_MONITOR_INFO\n", dev->name) );
		
		// Hum, seem that 'len' is ALWAYS zero !!!
		// if ( len < sizeof(*sm) )
		// 	 return EINVAL;
		
		sm = (SENSORS_MONITOR *) data;
		
		sm->api_version = SENSOR_CURRENT_API_VERSION;
		strncpy(sm->name, w->name, sizeof(sm->name));
		sm->nb_sensors = dev->nb_sensors;
		break;
		};

	case SENSOR_GET_VALUE:
		{
		SENSOR *	s;
		int			id;
		
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_GET_VALUE\n", dev->name) );

		// Hum, seem that 'len' is ALWAYS zero !!!
		// if ( len < sizeof(*s) )
		//   return EINVAL;
			
		s = (SENSOR *) data;
		id = s->id;
		if ( id < 0 || id > (dev->nb_sensors - 1) )
			return EINVAL;
			
		w8378xx_update_values(dev);
			
		memcpy(s, &(dev->sensors[id]), sizeof(*s));
		s->when = system_time();
		break;
		};

	case SENSOR_COMMAND:
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_COMMAND\n", dev->name) );
		return EINVAL;
		};

	default:
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: op_code %d (%x) unhandled.\n", dev->name, op_code, op_code) );
		return EINVAL;
		};
	}
	return B_OK;
}


// -------------------------------------------------------
static status_t read_hook
	(
	void *		cookie, 
	off_t 		pos, 
	void *		data, 
	size_t *	len
	)
{
	device *	dev;
	int			i;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- read_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- read_hook %s: %x, %d\n",
				dev->name, data, *len ) );
	
	/*
	for( i = 0; i < dev->nb_sensors; i++ )
		{
		sprintf( data, "%d\t[%s]\t%f\n",
				 dev->sensors[i].id, dev->sensors[i].name, dev->sensors[i].value ); 
		};
*/
	// *len = strlen(data);
	// return B_OK;

	*len = 0;
	return B_IO_ERROR;
}



// -------------------------------------------------------
static status_t write_hook
	(
	void *			cookie, 
	off_t 			pos, 
	const void *	data, 
	size_t *		len
	)
{
	device *	dev;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- write_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- write_hook %s: %x, %d\n",
				dev->name, data, *len ) );

	*len = 0;
	return B_IO_ERROR;
}


#if 0
	dprintf( "0x47+0x49 VID0-4 :\n" );
	isa->write_io_8(LM78_ADDRESS_REG, 0x47); // VID/Fan Divisor Register
	ret = isa->read_io_8(LM78_DATA_REG);	
	dprintf( "%2x+", ret );
	isa->write_io_8(LM78_ADDRESS_REG, 0x49); // Voltage ID
	ret = isa->read_io_8(LM78_DATA_REG);	
	dprintf( "%2x\n", ret );
	// Set FAN1 divisor to 4, and FAN2 divisor to 2
	LM78_WRITE(LM78_FANDIV, 0x60);
	// Set FAN3 to divide by 2
	dprintf( "0x4B Pin Control Register :\n" );
	isa->write_io_8(LM78_ADDRESS_REG, 0x4B); // Pin Control Register
	ret = isa->read_io_8(LM78_DATA_REG);	
	dprintf( "%2x\n", ret );
	ret &= 0x3F;
	ret |= 0x40;
	isa->write_io_8(LM78_ADDRESS_REG, 0x4B); // Pin Control Register
	isa->write_io_8(LM78_DATA_REG, ret);	
	// Verification
	isa->write_io_8(LM78_ADDRESS_REG, 0x4B); // Pin Control Register
	ret = isa->read_io_8(LM78_DATA_REG);	
	dprintf( "%2x\n", ret );
#endif


#ifdef CODEWARRIOR
	#pragma mark [Privates routines]
#endif

// Privates routines
// -----------------


// -------------------------------------------------------
static status_t device_find_supported
	(
	long * 				i,
	void ** 			ddd		// for "detected_device_data"
	)
{
	status_t	status;
	W8378xx *	w;
	long		n;

/*
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- device_find_supported(%x (-> %d), %x)\n",
				i, ((i == NULL) ? 0 : *i), ds) );
*/
	if ( NULL == g_isa )
		{
		status = get_module(B_ISA_MODULE_NAME, (module_info **) &g_isa);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": get_module(\"" B_ISA_MODULE_NAME "\") FAILED: %s\n",
							strerror(status)) );
			return ENOSYS;
			};
		};

	if ( NULL == i )
		return B_ERROR;
		
	// Okay, skip already tried entries	

	w = g_w8378xx_supported;
	n = *i;
	while ( n )
		{
		if ( w->name == NULL )
			return B_BAD_INDEX;
			
		n++;
		w++;
		};	
		
	// Detection of w8378xx chip(s) :
	// We try on w8378xx chip at default base IO port :
	while ( w->name != NULL )
		{
		uint16	vendor_id;
		uint8	ret;
		
		// We check Winbond Vendor ID register 

		// First, the High Byte of W8378xx_REG_WINBOND_VENDOR_ID register
		w8378xx_write(w, W8378xx_REG_BANK_SELECT, 0x80); 	// HBACS on, bank 0
		ret = w8378xx_read(w, W8378xx_REG_WINBOND_VENDOR_ID);
		vendor_id = (ret << 8);

		// Now, the Low Byte part
		w8378xx_write(w, W8378xx_REG_BANK_SELECT, 0x00); 	// HBACS off, bank 0
		ret = w8378xx_read(w, W8378xx_REG_WINBOND_VENDOR_ID);
		vendor_id |= ret;

		// dprintf(DRIVER_NAME ": vendor_id 0x%04x.\n", vendor_id);

		if ( vendor_id == WINBOND_VENDOR_ID )
			{
			// Wooooh, great star!
			// Okay, but which kind of w8378xx chip?
			uint8	chip_id;

			chip_id = w8378xx_read(w, W8378xx_REG_CHIP_ID);
			// dprintf(DRIVER_NAME ": chip_id 0x%x.\n", chip_id );

			chip_id &= W8378xx_CHIP_MASK;
			// dprintf(DRIVER_NAME ": chip_id masked 0x%x.\n", chip_id );

			if ( (chip_id & W8378xx_CHIP_MASK) == w->chip_id )
				{
				DEBUG_ONLY( dprintf(DRIVER_NAME ": chip ID 0x%x, %s detected at 0x%x.\n",
									chip_id, w->name, w->io_base )	);
				*i 		= n;
				*ddd 	= w; 

				return B_OK;
				};
			};
		
		n++;
		w++;
		// Perhaps next supported entry will be THE one ?
		};

	return B_BAD_INDEX;
}


// -------------------------------------------------------
static status_t device_reset
	(
	void *	detected_device_data
	)
{
	W8378xx *	w;
	
	w = detected_device_data;

	return B_OK;		
}


// -------------------------------------------------------
static status_t device_init
	(
	device *	dev
	)
{
	uint32			size;
	int 			i;
	status_t		status;

	status = B_ERROR;
	
	if ( NULL == dev )
		{
		status = B_BAD_VALUE;
		goto EXIT;
		};
		
	if ( NULL == g_isa )
		{
		status = ENOSYS;
		goto EXIT;
		};

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- device_init %s\n", dev->name) );
		
	dev->hardware_spinlock 	= 0;

	dev->update_semaphore	= create_sem(1, DRIVER_NAME "_update_semaphore");
	set_sem_owner(dev->update_semaphore, B_SYSTEM_TEAM); // Make kernel owner of update_semaphore

	dev->last_updated		= 0;	// not yet updated values
	
	status = w8378xx_init(dev);

EXIT:;

	delete_sem(dev->init_semaphore); // kill / release concurrent device_init()
	dev->init_semaphore = -1;
	
	return status;
}


// -------------------------------------------------------
static status_t device_uninit
	(
	device *	dev
	)
{
	w8378xx_uninit(dev);

	delete_sem(dev->update_semaphore);
	dev->update_semaphore = -1;

	return B_OK;
}



#ifdef CODEWARRIOR
	#pragma mark [Debugger Command]
#endif


#ifdef DEBUGGER_COMMAND

// -------------------------------------------------------
static int device_debugger_command
	(
	int 	argc,
	char **	argv
	)
{
	struct debugger_commands
		{
		int		id;
		char * 	name;
		char *	shortcut;
		// TO DO : command handling by sub-routine call
		// int 	(*cmd)(device * dev, int argc, char ** argv);
		char *	arguments;
		}
	* cmd, dbgr_cmds[] =
		{
			{0,  "help", "h", NULL},
			{1,  "list", "l", NULL},
			{2,  "select", "s", "{index}"},
			{3,  "stats", "st", NULL},
			{4,  "clear_stats", "cs", NULL},
			{5,  "dump_registers", "dr", NULL},
			{6, "lockers", "lck", NULL},
			{0, NULL, NULL, NULL}
		};
	int 			cmd_id;
	int				return_value;
	int				i;
	W8378xx *		w;
	
	static int		dev_id = 0;
	static device *	dev;

	return_value 	= 0;
	w 				= NULL;
	dev 			= g_device_list[dev_id];
	if ( NULL == dev )
		kprintf("  no device selected !\n");
	else
		w = dev->data;

	cmd_id 			= 0;	// help, by default

	if ( argc > 1 )
		{
		cmd = dbgr_cmds;
		while ( cmd->name )
			{
			if ( (strcmp(cmd->name, argv[1])== 0) ||
				 (strcmp(cmd->shortcut, argv[1]) == 0 ) )
				break;

			 cmd++;
			};

		cmd_id = cmd->id;
		};

	switch (cmd_id)
		{
		case 1: // list
			{
			for ( i = 0; g_device_list[i]; i++ )
				{
				kprintf("  %d: %s", i, g_device_list[i]->name);
				if ( i = dev_id )
					kprintf(" (selected)");
				kprintf("\n");
				};
			break;
			};

		case 2:	// select
			{
			if ( argc >= 3 )
				{
				i = atoi(argv[2]);
				if ( (i < 0) || (i >= DEVICE_MAX_ENTRIES) )
				 	kprintf("  index must be in [0, %d] range !\n", DEVICE_MAX_ENTRIES - 1);
				else
					dev_id = i;
				}
			else
				kprintf("  index missing !\n");
			break;
			};

		case 6: // dump_dev
			{
			if ( NULL == dev )
				break;

			kprintf("  device at 0x%4.4x :\n"
			        "    id                %d\n"
					"    name              %s\n"
					"    open_count        %d\n"
					"    flags             %x\n"
					"    chip              %s\n"
					"    base              0x%x\n",
					dev, dev->id, dev->name, dev->open_count, dev->flags,
					w->name, w->io_base);

			DEBUG_ONLY( kprintf("    debug_level       0x%x\n", dev->debug_level) );
			break;
			};

		case 9: // dump_registers
			{
			if ( NULL == dev )
				break;
			// TO DO : macro DBGR_READ_n(dev, reg)

/*
			kprintf("  registers :\n"
			        "    VR_PAR       ");
			for(i = 0; i < 5; i++)
				kprintf("0x1.1x", HW_READ_8(dev, VR_PAR0 + i);
			kprintf("\n";
			
			kprintf("    VR_RX_CFG        0x%x\n"
			        "    VR_TX_CFG        0x%x\n"
			        "    VR_CMD           0x%x\n
			        "    VR_ISR           0x%x\n"
			        "    VR_IMR           0x%x\n"
			        "    VR_MAR0          %d\n"
			        "    VR_MAR1          %d\n",
					HW_READ_8(dev, VR_RX_CFG), 
					HW_READ_8(dev, VR_TX_CFG), 
					HW_READ_32(dev, VR_CMD), 
					HW_READ_16(dev, VR_ISR), 
					HW_READ_16(dev, VR_IMR), 
					HW_READ_32(dev, VR_MAR0),
					HW_READ_32(dev, VR_MAR1) );
					);
*/

			break;
			};
			
		case 0:	// help
		default:
			{
			kprintf(DEBUGGER_COMMAND " usage:\n");
			cmd = dbgr_cmds;
			while ( cmd->name )
				{
				kprintf("  %s", cmd->name);
				if ( cmd->shortcut )
					kprintf(" | %s", cmd->shortcut);
				if ( cmd->arguments )
					kprintf(" %s", cmd->arguments);
				kprintf("\n");

				cmd++;	// next
				};
			
			break;
			};
		};	// switch
	
	return return_value;
}

#endif // DEBUGGER_COMMAND

#ifdef CODEWARRIOR
	#pragma mark [Tools]
#endif


// -------------------------------------------------------
static SENSOR *	device_find_sensor
	(
	device *	dev,	
	char *		name,
	int			id
	)
{
	int			i;

	if ( NULL == dev )
		return NULL;
		
	if ( NULL == dev->sensors )
		return NULL;

	for( i = 0; i < dev->nb_sensors; i++ )
		{
		if ( (name != NULL) && (strcmp(name, dev->sensors[i].name) == 0) )
			return &(dev->sensors[i]);
			
		if ( (name == NULL) && (dev->sensors[i].id == id) )
			return &(dev->sensors[i]);
		};

	return NULL;		
}


#ifdef CODEWARRIOR
	#pragma mark [w8378xx tools]
#endif

// -------------------------------------------------------
static int w8378xx_read
	(
	W8378xx * 	w,
	int			reg
	)
{
	int 			ret;
	cpu_status		cs;					// Tempory storage for disable_interrupt()
	spinlock		hardware_spinlock;	// Hardware access *hardlock*

	if ( w == NULL )
		return B_BAD_VALUE;
		
	if ( g_isa == NULL )
		return ENOSYS;
		
	cs = disable_interrupts();
	hardware_spinlock = 0;
	acquire_spinlock(&hardware_spinlock);		

	while ( g_isa->read_io_8(w->io_base + W8378xx_ADDRESS) & W8378xx_BUSY )
		spin(W8378xx_BUSY_WAIT);

	if ( w->current_reg_bank != (reg >> 8) )
		{
		// switch to another bank

		g_isa->write_io_8( w->io_base + W8378xx_ADDRESS, W8378xx_REG_BANK_SELECT ); // Bank Select
		g_isa->write_io_8( w->io_base + W8378xx_DATA, (reg >> 8));	                // Bank number

		// dprintf(DRIVER_NAME ": ---- w8378xx_read(0x%x) : switch to reg bank %d.\n", reg, (reg >> 8) );

		w->current_reg_bank = (reg >> 8);
		};

	g_isa->write_io_8( w->io_base + W8378xx_ADDRESS, (reg & 0xFF) ); 
	ret = g_isa->read_io_8( w->io_base + W8378xx_DATA);

	release_spinlock(&hardware_spinlock);
	restore_interrupts(cs);
	
	// dprintf(DRIVER_NAME ": ---- w8378xx_read(0x%x) -> %d (0x%x).\n", reg & 0xFF, ret, ret);

	return ret;
}	        


// -------------------------------------------------------
static status_t w8378xx_write
	(
	W8378xx * 	w,
	int			reg,
	int			value
	)
{
	cpu_status		cs;					// Tempory storage for disable_interrupt()
	spinlock		hardware_spinlock;	// Hardware access *hardlock*

	if ( w == NULL )
		return B_BAD_VALUE;
		
	if ( g_isa == NULL )
		return ENOSYS;
		
	cs = disable_interrupts();
	hardware_spinlock = 0;
	acquire_spinlock(&hardware_spinlock);		

	while ( g_isa->read_io_8(w->io_base + W8378xx_ADDRESS) & W8378xx_BUSY )
		spin(W8378xx_BUSY_WAIT);

	if ( w->current_reg_bank != (reg >> 8) )
		{
		// switch to another bank
		g_isa->write_io_8( w->io_base + W8378xx_ADDRESS, W8378xx_REG_BANK_SELECT ); // Bank Select
		g_isa->write_io_8( w->io_base + W8378xx_DATA, (reg >> 8));	                // Bank number

		// dprintf(DRIVER_NAME ": ---- w8378xx_write(0x%x, %d (0x%x) ) : switch to reg bank %d.\n", reg, value, value, (reg >> 8) );

		w->current_reg_bank = (reg >> 8);
		};

	g_isa->write_io_8( w->io_base + W8378xx_ADDRESS, (reg & 0xFF) ); 
	g_isa->write_io_8( w->io_base + W8378xx_DATA, value);

	release_spinlock(&hardware_spinlock);
	restore_interrupts(cs);

	// dprintf(DRIVER_NAME ": ---- w8378xx_write(0x%x, %d (0x%x)).\n", reg & 0xFF, value, value);
	
	return B_OK;
}	        


// -------------------------------------------------------
static status_t w8378xx_init
	(
	device *	dev
	)
{
	W8378xx * 			w;
	W8378xx_SENSORS * 	p;
	
	if ( (NULL == dev) || (dev->data == NULL) )
		return B_BAD_VALUE;

	// dprintf(DRIVER_NAME ": w8378xx_init: sensors list:\n");
	
	w = dev->data;

	// How many sensors ?
	dev->nb_sensors 	= 0;
	for ( p = g_w8378xx_sensors; p->name != NULL; p++ )
		{
		if ( p->unsupported & w->mask )
			// sensor unsupported by this chip
			continue;
		else
			dev->nb_sensors++;
		};	

	// Make room for sensors datas
	dev->sensors		= (SENSOR *) malloc((dev->nb_sensors * sizeof(SENSOR)) + 16);	

	// Load sensors descriptions & stuff
	dev->nb_sensors 	= 0;
	for ( p = g_w8378xx_sensors; p->name != NULL; p++ )
		{
		SENSOR * s;

		if ( p->unsupported & w->mask )
			// sensor unsupported by this chip
			continue;
			
		s = &(dev->sensors[dev->nb_sensors]);
			
		s->id 				= dev->nb_sensors;

		strncpy(s->name, p->name, sizeof(s->name));
		s->kind 			= p->kind;
		s->flags 			= p->flags;
		s->when				= 0;	// need to be updated !!!
		s->value 			= 0;
		s->value_multiplier = p->value_multiplier;
		s->value_divider	= p->value_divider;
				
		// dprintf(DRIVER_NAME ": %d\t%s\t%d\t%d\n",
		//	    	s->id, s->name, s->kind, s->value );
						
		dev->nb_sensors++;
		};

	return B_OK;
}


// -------------------------------------------------------
static status_t w8378xx_uninit
	(
	device * dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	if ( dev->sensors != NULL )
		free(dev->sensors);
	dev->sensors = NULL;

	return B_OK;
}


// -------------------------------------------------------
static status_t w8378xx_update_values
	(
	device * dev
	)
{
	status_t	status;
	W8378xx *	w;

	if ( NULL == dev )
		return B_BAD_VALUE;

	w = dev->data;
	
	// Okay, acquire update sem right now, or just wait the pending one finish update phase
	status = acquire_sem_etc( dev->update_semaphore, 1, B_RELATIVE_TIMEOUT, 0 );
	if( status == B_NO_ERROR )
		{
		bigtime_t	when;
		
		when = system_time();
		
		if( when - dev->last_updated > W8378xx_UPDATE_PERIOD )
			{
			SENSOR *	s;
			
			// Need a update
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- w8378xx_update_values\n") );
		
			// read registers values
			dev->last_updated = when;
			
			// Okay, values updating -> read the registers values

			// ---- VCore A (2V) ----
			if ( (s = device_find_sensor(dev, "in0", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN0);
				
			// ---- VCore B (2V) ----
			if ( (s = device_find_sensor(dev, "in1", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN1);
				
			// ---- 3.3V ----
			if ( (s = device_find_sensor(dev, "in2", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN2);
				
			// ---- 5V ----
			if ( (s = device_find_sensor(dev, "in3", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN3);
				
			// ---- 12V ----
			if ( (s = device_find_sensor(dev, "in4", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN4);
				
			// ---- -12V ----
			if ( (s = device_find_sensor(dev, "in5", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_MINUS_IN5);
				
			// ---- -5V ----
			if ( (s = device_find_sensor(dev, "in6", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_MINUS_IN6);
				
			// ---- Hum? ----
			if ( (s = device_find_sensor(dev, "in7", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN7);
				
			// ---- Hum?, but bis ----
			if ( (s = device_find_sensor(dev, "in8", 0)) != NULL )
				s->value = w8378xx_read(w, W8378xx_VALUE_IN8);
				
			// ---- Temperature 1 (motherboard, often) ----
			if ( (s = device_find_sensor(dev, "temp1", 0)) != NULL )
				s->value = 10 * w8378xx_read(w, W8378xx_VALUE_TEMPERATURE1);
				
			// ---- Temperature 2 (CPU, very often) ----
			if ( (s = device_find_sensor(dev, "temp2", 0)) != NULL )
				{
				s->value = 10 * w8378xx_read(w, W8378xx_VALUE_TEMPERATURE2);
				if ( w8378xx_read(w, W8378xx_VALUE_TEMPERATURE2 + 1) & 0x80 )
					s->value += 5;
				};
				
			// ---- Temperature 3 (second CPU or other stuffs, like disks) ----
			if ( (s = device_find_sensor(dev, "temp3", 0)) != NULL )
				{
				s->value = 10 * w8378xx_read(w, W8378xx_VALUE_TEMPERATURE3);
				if ( w8378xx_read(w, W8378xx_VALUE_TEMPERATURE3 + 1) & 0x80 )
					s->value += 5;
				};
				
			// ---- Fan 1 ----
			if ( (s = device_find_sensor(dev, "fan1", 0)) != NULL )
				{
				int	fan_divisor;

				fan_divisor	= ((w8378xx_read(w, W8378xx_REG_VID03_FAN12DIV) >> 4) & 0x03);
			  	if ( w->chip_id != W83781D )
					fan_divisor |= (( w8378xx_read(w, W8378xx_REG_VBAT) >> 3) & 0x04);

				s->value = w8378xx_read(w, W8378xx_VALUE_FAN1);
				s->value_divider = 1;

				if ( s->value == W8378xx_NO_FAN )
					{
					// Ohoh, fan disabled or overcount fan...
					s->value_divider = 0; // this time, say his's disabled

					// Can we auto adjust the fan divisor ?
					if ( (1<<fan_divisor) < ((w->chip_id == W83781D) ? 8 : 128) )
						{
						int tmp_reg;

						// Perhaps it's just because divisor is too small ?
						fan_divisor++;	// Let's try the next, bigger one !

						DEBUG_ONLY( dprintf( DRIVER_NAME ": auto adjusting %s divisor : trying at %d\n", s->name, (1 << fan_divisor)) );
						
						tmp_reg = w8378xx_read(w, W8378xx_REG_VID03_FAN12DIV);
						tmp_reg &= 0xCF;	// zero fandiv1 bits0-1
						tmp_reg |= ((fan_divisor & 0x03) << 4);
						w8378xx_write(w, W8378xx_REG_VID03_FAN12DIV, tmp_reg);
						if (w->chip_id != W83781D )
							{
							tmp_reg = w8378xx_read(w, W8378xx_REG_VBAT);
							tmp_reg &= 0xDF; // zero fandiv1 bit2
							tmp_reg |= ((fan_divisor & 0x04) << 3);
							};
						};
					}
				else
					s->value = W8378xx_FAN_COUNT_TO_RPM(s->value, 1 << fan_divisor );
				};
				
			// ---- Fan 2 ----
			if ( (s = device_find_sensor(dev, "fan2", 0)) != NULL )
				{
				int	fan_divisor;

				fan_divisor	= ((w8378xx_read(w, W8378xx_REG_VID03_FAN12DIV) >> 6) & 0x03);
			  	if ( w->chip_id != W83781D )
					fan_divisor |= (( w8378xx_read(w, W8378xx_REG_VBAT) >> 4) & 0x04);

				s->value = w8378xx_read(w, W8378xx_VALUE_FAN2);
				s->value_divider = 1;

//				if ( ! w8378xx_auto_adjust_fan_divisor(s, fan_divisor) )
// W8378xx_FAN_OVERRUN
// W8378xx_FAN_UNDERRUN

				if ( s->value < W8378xx_FAN_OVERRUN )
					
				if ( s->value == W8378xx_NO_FAN )
					{
					// Ohoh, fan disabled or overcount fan...
					s->value_divider = 0; // this time, say his's disabled

					// Can we auto adjust the fan divisor ?
					if ( (1<<fan_divisor) < ((w->chip_id == W83781D) ? 8 : 128) )
						{
						int tmp_reg;

						// Perhaps it's just because divisor is too small ?
						fan_divisor++;	// Let's try the next, bigger one !

						DEBUG_ONLY( dprintf( DRIVER_NAME ": auto adjusting %s divisor : trying at %d\n", s->name, (1 << fan_divisor)) );
						
						tmp_reg = w8378xx_read(w, W8378xx_REG_VID03_FAN12DIV);
						tmp_reg &= 0x3F;	// zero fandiv2 bits0-1
						tmp_reg |= ((fan_divisor & 0x03) << 6);
						w8378xx_write(w, W8378xx_REG_VID03_FAN12DIV, tmp_reg);
						if (w->chip_id != W83781D )
							{
							tmp_reg = w8378xx_read(w, W8378xx_REG_VBAT);
							tmp_reg &= 0xBF; // zero fandiv2 bit2
							tmp_reg |= ((fan_divisor & 0x04) << 4);
							};
						};
					}
				else
					if ( s->value != 0 )
						s->value = W8378xx_FAN_COUNT_TO_RPM(s->value, 1 << fan_divisor );
				};
				
			// ---- Fan 3 ----
			if ( (s = device_find_sensor(dev, "fan3", 0)) != NULL )
				{
				int	fan_divisor;

				fan_divisor = ((w8378xx_read(w, W8378xx_REG_FAN3DIV_PIN_CONTROL) >> 6) & 0x03);
			  	if ( w->chip_id != W83781D )
					fan_divisor |= (( w8378xx_read(w, W8378xx_REG_VBAT) >> 5) & 0x04);

				s->value = w8378xx_read(w, 	W8378xx_VALUE_FAN3);
				s->value_divider = 1;

				if ( s->value == W8378xx_NO_FAN )
					{
					// Ohoh, fan disabled or overcount fan...
					s->value_divider = 0; // this time, say his's disabled

					// Can we auto adjust the fan divisor ?
					if ( (1<<fan_divisor) < ((w->chip_id == W83781D) ? 8 : 128) )
						{
						int tmp_reg;

						// Perhaps it's just because divisor is too small ?
						fan_divisor++;	// Let's try the next, bigger one !

						DEBUG_ONLY( dprintf( DRIVER_NAME ": auto adjusting %s divisor : trying at %d\n", s->name, (1 << fan_divisor)) );
						
						tmp_reg = w8378xx_read(w, W8378xx_REG_FAN3DIV_PIN_CONTROL);
						tmp_reg &= 0x3F;	// zero fandiv3 bits0-1
						tmp_reg |= ((fan_divisor & 0x03) << 6);
						w8378xx_write(w, W8378xx_REG_FAN3DIV_PIN_CONTROL, tmp_reg);
						if (w->chip_id != W83781D )
							{
							tmp_reg = w8378xx_read(w, W8378xx_REG_VBAT);
							tmp_reg &= 0x7F; // zero fandiv3 bit2
							tmp_reg |= ((fan_divisor & 0x04) << 5);
							};
						};
					}
				else
					if ( s->value != 0 )
						s->value = W8378xx_FAN_COUNT_TO_RPM(s->value, 1 << fan_divisor );
				};

			// ---- Voltage ID ----
			if ( (s = device_find_sensor(dev, "vid", 0)) != NULL )
				{
				s->value = 	(w8378xx_read(w, W8378xx_REG_VID03_FAN12DIV) & 0x0f );
				s->value |=	((w8378xx_read(w, W8378xx_REG_VID4_ID) & 0x01) << 4);
				
				s->value = g_w8378xx_VID[s->value];
				if ( s->value == 0 )
					// Not supported VID...
					s->value_divider = 0;
				};
				
#if 0	
			// Value RAM (auto-increment)
			g_isa->write_io_8( W8378xx_IO_ADDRESS, W8378xx_VALUE_RAM_AUTOINDEX );	
			
			dprintf( "0x60 IN0 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x61 IN1 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x62 IN2 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x63 IN3 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x64 IN4 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x65 -IN5 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x66 -IN6 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x67 TEMP1 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x68 FAN1 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x69 FAN2 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6A FAN3 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6B IN0HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6C IN0LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6D IN1HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6E IN1LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x6F IN2HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x70 IN2LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x71 IN3HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x72 IN3LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x73 IN4HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x74 IN4LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x75 -IN5HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x76 -IN5LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x77 -IN6HL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x78 -IN6LL = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x79 OVERTEMP1 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7A HYSTERTEMP1 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7B FAN1LIMIT = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7C FAN2LIMIT = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7D FAN3LIMIT = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7E = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			dprintf( "0x7F = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			
			// Read extended sensors : temp2 & temp3 
			// temp2 is on bank 1
			g_isa->write_io_8( W8378xx_IO_ADDRESS, W8378xx_REG_BANK_SELECT ); // Bank Select
			g_isa->write_io_8( W8378xx_IO_DATA, 0x01);	                   // HBACS off, bank 1
			g_isa->write_io_8( W8378xx_IO_ADDRESS, 0x50); // Temp2
			dprintf( "0x150 TEMP2 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			g_isa->write_io_8( W8378xx_IO_ADDRESS, 0x51); // Temp2 low digit
			dprintf( "0x151 TEMP2 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
	
			// temp3 is on bank 2
			g_isa->write_io_8( W8378xx_IO_ADDRESS, W8378xx_REG_BANK_SELECT ); // Bank Select
			g_isa->write_io_8( W8378xx_IO_DATA, 0x02);	                   // HBACS off, bank 2
			g_isa->write_io_8( W8378xx_IO_ADDRESS, 0x50); // Temp3
			dprintf( "0x250 TEMP3 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
			g_isa->write_io_8( W8378xx_IO_ADDRESS, 0x51); // Temp3 low digit
			dprintf( "0x251 TEMP3 = %d\n", g_isa->read_io_8( W8378xx_IO_DATA ) );
	
			// Return to bank 0 by default
			g_isa->write_io_8( W8378xx_IO_ADDRESS, W8378xx_REG_BANK_SELECT ); // Bank Select
			g_isa->write_io_8( W8378xx_IO_DATA, 0x00);	                   // HBACS off, bank 1
	
#endif	
			dev->nb_updates++;
			// Okay, it's done.
			};
		}
	else if ( status = B_WOULD_BLOCK )
		// okay, concurrent thread run the update process, just wait it's end
		acquire_sem(dev->update_semaphore);
	else
		// 	Argh, something going wrong with update semaphore !
		return status;
		 
	// Update is done, by this thread or a concurrent thread, but who mind ? 
	release_sem( dev->update_semaphore );
	
	return B_OK;
}



